
      !-----------------------------------------------------------------
	!
      !  Purpose: To compute geopotential acceleration
	!
	!  Inputs:
	!     NMAX  I2 max. order and degree loaded
	!       Nm  I2 desired order and degree (Nm <= NMAX)
	!       Ae  R8 mean equatorial radius of the model [m]
	!       GM  R8 Earth gravitational constant [m**3/s**2]
	!        C  R8 array (0:NMAX,0:Nm) Harmonic coefficients C(n,m)
	!        S  R8 array (0:NMAX,0:Nm) Harmonic coefficients S(n,m)
	!        X  R8 array (3) ECEF cartesian coordinates [m]         
	!
	!  Outputs:
	!       Ac  R8 array (3) with geopotential (x, y, z) accelerations [m/s**2]
	!
	!  Author: HKK - INPE - May 2012 - Version 1.0
	!
      !  Refs.: 
	!
	!    Kuga, H.K.; Carrara, V. "Fortran- and C-codes for higher order
	!      and degree geopotential computation." 
	!      www.dem.inpe.br/~hkk/software/high_geopot.html
	!      Last access May 2012.
	!
      !-----------------------------------------------------------------

      subroutine Leg_ForCol_Ac (NMAX, Nm, Ae, GM, C, S, X, Ac)

	implicit none

	integer*2 NMAX, Nm
	REAL*8 Ae, GM, C(0:NMAX,0:Nm), S(0:NMAX,0:Nm), X(3), Ac(3)

      ! Locals
      integer*2 n, m
      real*8 r, q, t, u2, u, um, tf, al, sl, cl, Gmr
	real*8 pn(0:Nm), qn(0:Nm)
      real*8 Pnm, dPnm, anm, bnm, fnm
      real*8 am, an, Pnn, Pnm1m, Pnm2m, sm, cm, sml, cml
      real*8 qC, qS, Xc, Xs, Xcf, Xsf, Xcr, Xsr, Vl, Vf, Vr
!	real*8 Omega

      ! Auxiliary variables
      r = sqrt(x(1)*x(1) + x(2)*x(2) + x(3)*x(3))
	q = Ae / r
	t = x(3) / r  ! sin (lat)
	u = sqrt (1.d0 - t*t)
	tf = t/u        ! tan (lat)
	al = atan2(x(2), x(1))
	sl = sin (al)            ! sin (long)
      cl = cos (al)            ! cos (long)
	Gmr = GM / r
      
      ! Summation initialization
!      Omega = 0.d0
	Vl = 0.d0
	Vf = 0.d0
	Vr = 0.d0

      ! Store sectoral
	pn(0) = 1.d0
	pn(1) = 1.7320508075688773D0 * u ! sqrt(3) * cos (lat)
	qn(0) = 1.d0
	qn(1) = q
      do m = 2, Nm
	  am = dble(m)
        pn(m) = u * sqrt(1.d0+0.5d0/am) * pn(m-1)
	  qn(m) = q * qn(m-1)
      end do

      ! Initialize sin and cos recursions
      sm = 0.d0
	cm = 1.d0

      ! Outer loop
      do m=0,Nm
      
        ! init 
	  am = dble(m)

	  ! For m = n 
        Pnm = pn(m)             ! m=n sectoral
	  dPnm = -am * Pnm * tf
	  Pnm1m = Pnm
	  Pnm2m = 0.d0

        ! Init  Horner's scheme
        qC = qn(m) * C(m,m)
	  qS = qn(m) * S(m,m)
        Xc = qC * Pnm 
        Xs = qS * Pnm
        Xcf = qC * dPnm 
        Xsf = qS * dPnm 
	  Xcr = (am+1.d0) * qC * Pnm 
	  Xsr = (am+1.d0) * qS * Pnm 

        ! Inner Loop 
        do n=m+1,Nm
	    an = dble(n)
          anm = sqrt( ((an+an-1.d0)*(an+an+1.d0))
     *              / ((an-am)*(an+am)) )
	    bnm = sqrt( ((an+an+1.d0)*(an+am-1.d0)*(an-am-1.d0))
     &              / ((an-am)*(an+am)*(an+an-3.d0)) )
	    fnm = sqrt( ((an*an-am*am)*(an+an+1.d0))/(an+an-1.d0) )
          ! recursion P and dP
          Pnm = anm * t * Pnm1m - bnm * Pnm2m
          dPnm = -an * Pnm * tf + fnm * Pnm1m / u   ! Signal opposite to paper

          ! Store
	    Pnm2m = Pnm1m
	    Pnm1m = Pnm

          ! Inner sum
          if (n .lt. 2) cycle
          qC = qn(n) * C(n,m)
	    qS = qn(n) * S(n,m)
          Xc = (Xc + qC * Pnm)
          Xs = (Xs + qS * Pnm)
          Xcf = (Xcf + qC * dPnm)
          Xsf = (Xsf + qS * dPnm)
	    Xcr = (Xcr + (an+1.d0) * qC * Pnm) 
	    Xsr = (Xsr + (an+1.d0) * qS * Pnm) 
	  end do

        ! Outer sum
! 	  Omega = Omega + (Xc*cm + Xs*sm)  
        Vl = Vl + am*(Xc*sm - Xs*cm) 
        Vf = Vf + (Xcf*cm + Xsf*sm) 
        Vr = Vr + (Xcr*cm + Xsr*sm)

        ! sin and cos recursions to next m
	  cml = cl*cm - sm*sl
        sml = cl*sm + cm*sl
        cm = cml ! save to next m
	  sm = sml ! save to next m

      end do

      ! Finalization, include n=0 (P00=1), 
	! for n=1 all terms are zero: C,S(1,1), C,S(1,0) = 0

	! Potential
!      Omega =  Gmr * (1.d0+Omega)

	! Gradient 
      Vl = -Gmr * Vl
      Vf =  Gmr * Vf
      Vr = -(Gmr/r) * (1.d0+Vr)

      ! Body X, Y, Z accelerations 
      Ac(1) = u*cl*Vr - t*cl*Vf/r - sl*Vl/(u*r)
	Ac(2) = u*sl*Vr - t*sl*Vf/r + cl*Vl/(u*r)
	Ac(3) = t*Vr + u*Vf/r

      return
	end