      !-----------------------------------------------------------------
	!
      !  Purpose: To compute geopotential
	!
	!  Inputs:
	!     NMAX  I2 max. order and degree loaded
	!       Nm  I2 desired order and degree (Nm <= NMAX)
	!       Ae  R8 mean equatorial radius of the model [m]
	!       GM  R8 Earth gravitational constant [m**3/s**2]
	!        C  R8 array (0:NMAX,0:Nm) Harmonic coefficients C(n,m)
	!        S  R8 array (0:NMAX,0:Nm) Harmonic coefficients S(n,m)
	!        X  R8 array (3) ECEF cartesian coordinates [m]         
	!
	!  Outputs:
	!        V  R8 geopotential value [m**2/s**2]
	!
	!  Author: HKK - INPE - May 2012 - Version 1.0
	!
      !  Refs.: 
	!
	!    Kuga, H.K.; Carrara, V. "Fortran- and C-codes for higher order
	!      and degree geopotential computation." 
	!      www.dem.inpe.br/~hkk/software/high_geopot.html
	!      Last access May 2012.
	!
      !-----------------------------------------------------------------

      subroutine Leg_ForCol_Pot (NMAX, Nm, Ae, GM, C, S, X, V)

	implicit none

      ! Calling parameters
	integer*2 NMAX, Nm
	real*8 Ae, GM, C(0:NMAX,0:Nm), S(0:NMAX,0:Nm), X(3), V

      ! Locals
      integer*2 n, m
      real*8 r, q, t, u, tf, al, sl, cl, Gmr
	real*8 pn(0:Nm), qn(0:Nm)
      real*8 Pnm, anm, bnm
      real*8 am, an, Pnn, Pnm1m, Pnm2m, sm, cm, sml, cml
      real*8 Xc, Xs, Omega

      ! Auxiliary variables
      r = sqrt(x(1)*x(1) + x(2)*x(2) + x(3)*x(3))
	q = Ae / r
	t = x(3) / r           ! sin (lat)
	u = sqrt (1.d0 - t*t)  ! cos (lat)
	tf = t/u               ! tan (lat)
	al = atan2(x(2), x(1)) ! longitude
	sl = sin (al)          ! sin (long)
      cl = cos (al)          ! cos (long)
	Gmr = GM / r
      
      ! Summation initialization
      Omega = 0.d0

      ! Pre-Store sectoral polynomials and q(m) = (Ae/r)**m
	pn(0) = 1.d0
	pn(1) = 1.7320508075688773D0 * u ! sqrt(3) * u
	qn(0) = 1.d0
	qn(1) = q
      do m = 2, Nm
        pn(m) = u * sqrt(1.d0+0.5d0/dble(m)) * pn(m-1)
	  qn(m) = q * qn(m-1)
      end do

      ! Initialize sin and cos recursions
      sm = 0.d0
	cm = 1.d0

      ! Outer m loop
      do m=0,Nm
      
        ! init 
	  am = dble(m)

	  ! For m = n 
        Pnm = pn(m)             ! m=n sectoral
	  Pnm1m = Pnm
	  Pnm2m = 0.d0

        ! Init Xc, Xs sums
        Xc = qn(m) * C(m,m) * Pnm 
        Xs = qn(m) * S(m,m) * Pnm

        ! Inner n Loop 
        do n=m+1,Nm
	    an = dble(n)
          anm = sqrt( ((an+an-1.d0)*(an+an+1.d0))
     &              / ((an-am)*(an+am)) )
	    bnm = sqrt( ((an+an+1.d0)*(an+am-1.d0)*(an-am-1.d0))
     &              / ((an-am)*(an+am)*(an+an-3.d0)) )

          ! Pnm recursion 
          Pnm = anm * t * Pnm1m - bnm * Pnm2m

          ! Store
	    Pnm2m = Pnm1m
	    Pnm1m = Pnm

          ! Inner sum
          if (n .lt. 2) cycle
          Xc = Xc + qn(n) * C(n,m) * Pnm
          Xs = Xs + qn(n) * S(n,m) * Pnm

	  end do

        ! Outer sum
	  Omega = Omega + (Xc*cm + Xs*sm) 

        ! sin and cos recursions to next m
	  cml = cl*cm - sm*sl
        sml = cl*sm + cm*sl
        cm = cml ! save to next m
	  sm = sml ! save to next m

      end do

      ! Finalization, include n=0 (P00=1), 
	! for n=1 all terms are zero: C,S(1,1), C,S(1,0) = 0
      V =  Gmr * (1.d0+Omega)

      return
	end